TEST_CERT_DIRECTORY="../pems/"
TEST_SNI_CERT_DIRECTORY="../pems/sni/"

TRUST_STORE_BUNDLE=TEST_CERT_DIRECTORY + 'trust-store/ca-bundle.crt'
TRUST_STORE_TRUSTED_BUNDLE=TEST_CERT_DIRECTORY + 'trust-store/ca-bundle.trust.crt'
