/**
 * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0.
 *
 * Empty source file. This is required because some proofs, like
 * memset_using_uint64, do not link against any project files at all. This file
 * is used to fill in the PROJECT_SOURCES array in such cases.
 */
